var EPSILON = Number.EPSILON || 2.2204460492503130808472633361816E-16;
function createAkimaSplineInterpolator(xvals, yvals) {
    var MINIMUM_NUMBER_POINTS = 5;
    var n = xvals.length;
    if (n != yvals.length) {
        throw new Error("Dimension mismatch for xvals and yvals.");
    }
    if (n < MINIMUM_NUMBER_POINTS) {
        throw new Error("Number of points is too small.");
    }
    MathArrays_checkOrder(xvals);
    var numberOfDiffAndWeightElements = n - 1;
    var differences = Array(numberOfDiffAndWeightElements);
    var weights = Array(numberOfDiffAndWeightElements);
    for (var i = 0; i < differences.length; i++) {
        differences[i] = (yvals[i + 1] - yvals[i]) / (xvals[i + 1] - xvals[i]);
    }
    for (var i = 1; i < weights.length; i++) {
        weights[i] = Math.abs(differences[i] - differences[i - 1]);
    }
    var firstDerivatives = Array(n);
    for (var i = 2; i < n - 2; i++) {
        var wP = weights[i + 1];
        var wM = weights[i - 1];
        if (Math.abs(wP) < EPSILON && Math.abs(wM) < EPSILON) {
            var xv = xvals[i];
            var xvP = xvals[i + 1];
            var xvM = xvals[i - 1];
            firstDerivatives[i] = (((xvP - xv) * differences[i - 1]) + ((xv - xvM) * differences[i])) / (xvP - xvM);
        }
        else {
            firstDerivatives[i] = ((wP * differences[i - 1]) + (wM * differences[i])) / (wP + wM);
        }
    }
    firstDerivatives[0] = differentiateThreePoint(xvals, yvals, 0, 0, 1, 2);
    firstDerivatives[1] = differentiateThreePoint(xvals, yvals, 1, 0, 1, 2);
    firstDerivatives[n - 2] = differentiateThreePoint(xvals, yvals, n - 2, n - 3, n - 2, n - 1);
    firstDerivatives[n - 1] = differentiateThreePoint(xvals, yvals, n - 1, n - 3, n - 2, n - 1);
    return interpolateHermiteSorted(xvals, yvals, firstDerivatives);
}
function differentiateThreePoint(xvals, yvals, indexOfDifferentiation, indexOfFirstSample, indexOfSecondsample, indexOfThirdSample) {
    var x0 = yvals[indexOfFirstSample];
    var x1 = yvals[indexOfSecondsample];
    var x2 = yvals[indexOfThirdSample];
    var t = xvals[indexOfDifferentiation] - xvals[indexOfFirstSample];
    var t1 = xvals[indexOfSecondsample] - xvals[indexOfFirstSample];
    var t2 = xvals[indexOfThirdSample] - xvals[indexOfFirstSample];
    var a = (x2 - x0 - (t2 / t1 * (x1 - x0))) / (t2 * t2 - t1 * t2);
    var b = (x1 - x0 - a * t1 * t1) / t1;
    return (2 * a * t) + b;
}
function interpolateHermiteSorted(xvals, yvals, firstDerivatives) {
    var minimumLength = 2;
    var n = xvals.length;
    if (n != yvals.length || n != firstDerivatives.length) {
        throw new Error("Dimension mismatch");
    }
    if (n < minimumLength) {
        throw new Error("Not enough points.");
    }
    var polynomials = Array(n - 1);
    var coefficients = Array(4);
    for (var i = 0; i < n - 1; i++) {
        var w = xvals[i + 1] - xvals[i];
        var w2 = w * w;
        var yv = yvals[i];
        var yvP = yvals[i + 1];
        var fd = firstDerivatives[i];
        var fdP = firstDerivatives[i + 1];
        coefficients[0] = yv;
        coefficients[1] = firstDerivatives[i];
        coefficients[2] = (3 * (yvP - yv) / w - 2 * fd - fdP) / w;
        coefficients[3] = (2 * (yv - yvP) / w + fd + fdP) / w2;
        polynomials[i] = createPolynomialFunction(coefficients);
    }
    return createPolynomialSplineFunction(xvals, polynomials);
}
function createCubicSplineInterpolator(x, y) {
    if (x.length != y.length) {
        throw new Error("Dimension mismatch.");
    }
    if (x.length < 3) {
        throw new Error("Number of points is too small.");
    }
    var n = x.length - 1;
    MathArrays_checkOrder(x);
    var h = Array(n);
    for (var i = 0; i < n; i++) {
        h[i] = x[i + 1] - x[i];
    }
    var mu = Array(n);
    var z = Array(n + 1);
    mu[0] = 0;
    z[0] = 0;
    var g = 0;
    for (var i = 1; i < n; i++) {
        g = 2 * (x[i + 1] - x[i - 1]) - h[i - 1] * mu[i - 1];
        mu[i] = h[i] / g;
        z[i] = (3 * (y[i + 1] * h[i - 1] - y[i] * (x[i + 1] - x[i - 1]) + y[i - 1] * h[i]) /
            (h[i - 1] * h[i]) - h[i - 1] * z[i - 1]) / g;
    }
    var b = Array(n);
    var c = Array(n + 1);
    var d = Array(n);
    z[n] = 0;
    c[n] = 0;
    for (var j = n - 1; j >= 0; j--) {
        c[j] = z[j] - mu[j] * c[j + 1];
        b[j] = (y[j + 1] - y[j]) / h[j] - h[j] * (c[j + 1] + 2 * c[j]) / 3;
        d[j] = (c[j + 1] - c[j]) / (3 * h[j]);
    }
    var polynomials = Array(n);
    var coefficients = Array(4);
    for (var i = 0; i < n; i++) {
        coefficients[0] = y[i];
        coefficients[1] = b[i];
        coefficients[2] = c[i];
        coefficients[3] = d[i];
        polynomials[i] = createPolynomialFunction(coefficients);
    }
    return createPolynomialSplineFunction(x, polynomials);
}
function createLinearInterpolator(x, y) {
    if (x.length != y.length) {
        throw new Error("Dimension msmatch.");
    }
    if (x.length < 2) {
        throw new Error("Number of points is too small.");
    }
    var n = x.length - 1;
    MathArrays_checkOrder(x);
    var m = Array(n);
    for (var i = 0; i < n; i++) {
        m[i] = (y[i + 1] - y[i]) / (x[i + 1] - x[i]);
    }
    var polynomials = Array(n);
    var coefficients = Array(2);
    for (var i = 0; i < n; i++) {
        coefficients[0] = y[i];
        coefficients[1] = m[i];
        polynomials[i] = createPolynomialFunction(coefficients);
    }
    return createPolynomialSplineFunction(x, polynomials);
}
function createPolynomialSplineFunction(knots, polynomials) {
    knots = knots.slice();
    if (knots.length < 2) {
        throw new Error("Not enough knots.");
    }
    if (knots.length - 1 != polynomials.length) {
        throw new Error("Dimension mismatch.");
    }
    MathArrays_checkOrder(knots);
    return function (x) {
        var i = Arrays_binarySearch(knots, x);
        if (i < 0) {
            i = -i - 2;
        }
        i = Math.max(0, Math.min(i, polynomials.length - 1));
        return polynomials[i](x - knots[i]);
    };
}
function createPolynomialFunction(c) {
    c = c.slice();
    var n = c.length;
    if (n == 0) {
        throw new Error("Empty polynomials coefficients array");
    }
    while (n > 1 && c[n - 1] == 0) {
        n--;
    }
    return function (x) {
        var v = c[n - 1];
        for (var i = n - 2; i >= 0; i--) {
            v = x * v + c[i];
        }
        return v;
    };
}
function MathArrays_checkOrder(val) {
    var previous = val[0];
    var max = val.length;
    for (var index = 1; index < max; index++) {
        if (val[index] <= previous) {
            throw new Error("Non-monotonic sequence exception.");
        }
        previous = val[index];
    }
}
function Arrays_binarySearch(a, key) {
    var low = 0;
    var high = a.length - 1;
    while (low <= high) {
        var mid = (low + high) >>> 1;
        var midVal = a[mid];
        if (midVal < key)
            low = mid + 1;
        else if (midVal > key)
            high = mid - 1;
        else if (midVal == key)
            return mid;
        else
            throw new Error("Invalid number encountered in binary search.");
    }
    return -(low + 1);
}

export { createAkimaSplineInterpolator, createCubicSplineInterpolator, createLinearInterpolator };
